import abc
from collections.abc import Iterable
from typing import Any, override

import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.linalg.abc import Callback, LinearSolution, LinearSolver, Result
from liblaf.peach.linalg.system import LinearSystem

from ._types import JaxState, JaxStats

type Scalar = Float[Array, ""]
type Vector = Float[Array, " free"]


@tree.define
class JaxSolver(LinearSolver[JaxState, JaxStats]):
    from ._types import JaxState as State
    from ._types import JaxStats as Stats

    Solution = LinearSolution[JaxState, JaxStats]

    atol: float = 0.0
    max_steps: int | None = None
    rtol: float = 1e-5

    @override
    def _solve(
        self,
        system: LinearSystem,
        state: State,
        stats: Stats,
        *,
        callback: Callback[State, Stats] | None = None,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[State, Stats, Result]:
        if constraints:
            raise NotImplementedError
        if callback is not None:
            raise NotImplementedError
        assert system.matvec is not None
        state.params_flat, _info = self._wrapped(
            system.matvec, system.b_flat, state.params_flat, **self._options(system)
        )
        residual: Vector = system.matvec(state.params_flat) - system.b_flat
        residual_norm: Scalar = jnp.linalg.norm(residual)
        b_norm: Scalar = jnp.linalg.norm(system.b_flat)
        stats.residual_relative = residual_norm / b_norm
        result: Result
        if residual_norm <= self.atol + self.rtol * b_norm:
            result = Result.SUCCESS
        else:
            result = Result.UNKNOWN_ERROR
        return state, stats, result

    def _options(self, system: LinearSystem) -> dict[str, Any]:
        max_steps: int = (
            system.b_flat.size if self.max_steps is None else self.max_steps
        )
        return {
            "tol": self.rtol,
            "atol": self.atol,
            "maxiter": max_steps,
            "M": system.preconditioner,
        }

    @abc.abstractmethod
    def _wrapped(self, *args, **kwargs) -> tuple[Vector, Any]:
        raise NotImplementedError



from liblaf.peach import tree
from liblaf.peach.linalg.abc import State, Stats


@tree.define
class CompositeState(State):
    state: list[State] = tree.field(factory=list)


@tree.define
class CompositeStats(Stats):
    stats: list[Stats] = tree.field(factory=list)



import abc
from collections.abc import Callable, Iterable
from typing import Any, override

import jax.numpy as jnp
import numpy as np
import scipy
from jaxtyping import Array, Float

from liblaf import grapes
from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.linalg.abc import (
    Callback,
    LinearSolution,
    LinearSolver,
    Params,
    Result,
    SetupResult,
)
from liblaf.peach.linalg.system import LinearSystem

type Free = Float[Array, " free"]
type FreeNp = Float[np.ndarray, " free"]


@tree.define
class ScipySolver(LinearSolver):
    from ._types import ScipyState as State
    from ._types import ScipyStats as Stats

    Solution = LinearSolution[State, Stats]

    rtol: float = tree.field(default=1e-5, kw_only=True)
    atol: float = tree.field(default=0.0, kw_only=True)
    max_steps: int | None = tree.field(default=None, kw_only=True)

    @override
    def setup(
        self,
        system: LinearSystem,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[State, Stats]:
        params_flat: Free
        system, params_flat, constraints = system.flatten(
            params, constraints=constraints
        )
        state: ScipySolver.State = self.State(
            params_flat=params_flat, flat_def=system.flat_def
        )
        if self.jit:
            system = system.jit()
        if self.timer:
            system = system.timer()
        return SetupResult(system, constraints, state, self.Stats())

    @override
    def _solve(
        self,
        system: LinearSystem,
        state: State,
        stats: Stats,
        *,
        callback: Callback[State, Stats] | None = None,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[State, Stats, Result]:
        if constraints:
            raise NotImplementedError
        cb_wrapper: Callable = self._make_callback(callback, state, stats)
        lop: scipy.sparse.linalg.LinearOperator = _as_lop(system)
        x: FreeNp
        info: int
        x, info = self._wrapped(
            lop,
            system.b_flat,
            state.params_flat,
            callback=cb_wrapper,
            **self._options(system),
        )
        state.params_flat = jnp.asarray(x)
        stats.n_steps = len(grapes.get_timer(cb_wrapper))
        result: Result
        stats, result = self._finalize(info, stats)
        return state, stats, result

    def _make_callback(
        self, callback: Callback[State, Stats] | None, state: State, stats: Stats
    ) -> Callable:
        @grapes.timer(label=f"{self.name}.callback()")
        def wrapper(xk: FreeNp) -> None:
            if callback is None:
                return
            state.params_flat = jnp.asarray(xk)
            stats.n_steps = len(grapes.get_timer(wrapper)) + 1
            callback(state, stats)

        return wrapper

    def _options(self, system: LinearSystem) -> dict[str, Any]:
        options: dict[str, Any] = {"rtol": self.rtol, "atol": self.atol}
        if self.max_steps is not None:
            options["maxiter"] = self.max_steps
        if system.preconditioner is not None:
            options["M"] = _preconditioner(system)
        return options

    def _finalize(self, info: int, stats: Stats) -> tuple[Stats, Result]:
        stats.info = info
        if info == 0:
            return stats, Result.SUCCESS
        if info < 0:
            return stats, Result.BREAKDOWN
        stats.n_steps = info
        return stats, Result.MAX_STEPS_REACHED

    @abc.abstractmethod
    def _wrapped(self, *args, **kwargs) -> tuple[FreeNp, int]:
        raise NotImplementedError


def _as_lop(system: LinearSystem) -> scipy.sparse.linalg.LinearOperator:
    assert system.matvec is not None

    def matvec(x: FreeNp) -> FreeNp:
        assert system.matvec is not None
        x_jax: Float[Array, " free"] = jnp.asarray(x)
        y_jax: Float[Array, " free"] = system.matvec(x_jax)
        return np.asarray(y_jax)

    def rmatvec(x: FreeNp) -> FreeNp:
        assert system.rmatvec is not None
        x_jax: Float[Array, " free"] = jnp.asarray(x)
        y_jax: Float[Array, " free"] = system.rmatvec(x_jax)
        return np.asarray(y_jax)

    dim: int
    (dim,) = system.b_flat.shape
    return scipy.sparse.linalg.LinearOperator(
        shape=(dim, dim),
        matvec=matvec,
        rmatvec=None if system.rmatvec is None else rmatvec,
        dtype=system.b_flat.dtype,
    )


def _preconditioner(system: LinearSystem) -> scipy.sparse.linalg.LinearOperator | None:
    if system.preconditioner is None:
        return None

    def matvec(x: Float[np.ndarray, " free"]) -> Float[np.ndarray, " free"]:
        assert system.preconditioner is not None
        x_jax: Float[Array, " free"] = jnp.asarray(x)
        y_jax: Float[Array, " free"] = system.preconditioner(x_jax)
        return np.asarray(y_jax)

    def rmatvec(x: Float[np.ndarray, " free"]) -> Float[np.ndarray, " free"]:
        assert system.rpreconditioner is not None
        x_jax: Float[Array, " free"] = jnp.asarray(x)
        y_jax: Float[Array, " free"] = system.rpreconditioner(x_jax)
        return np.asarray(y_jax)

    dim: int
    (dim,) = system.b_flat.shape
    return scipy.sparse.linalg.LinearOperator(
        shape=(dim, dim),
        matvec=matvec,
        rmatvec=None if system.rpreconditioner is None else rmatvec,
        dtype=system.b_flat.dtype,
    )



from . import abc, objective, pncg, scipy
from .abc import Callback, Optimizer, OptimizeSolution, Result
from .objective import Objective
from .pncg import PNCG
from .scipy import ScipyOptimizer

__all__ = [
    "PNCG",
    "Callback",
    "Objective",
    "OptimizeSolution",
    "Optimizer",
    "Result",
    "ScipyOptimizer",
    "abc",
    "objective",
    "pncg",
    "scipy",
]



import functools
from typing import Any, dataclass_transform

import attrs

from liblaf import grapes

from ._field import array, container, field
from ._register_fieldz import register_fieldz


@dataclass_transform(field_specifiers=(attrs.field, array, container, field))
@grapes.wraps(attrs.define)
def define(cls: type | None = None, /, **kwargs) -> Any:
    if cls is None:
        return functools.partial(define, **kwargs)
    cls = grapes.attrs.define(cls, **kwargs)
    cls = register_fieldz(cls)
    return cls



from ._base import ScipySolver
from ._bicg import ScipyBiCG
from ._bicgstab import ScipyBiCGStab
from ._cg import ScipyCG
from ._minres import ScipyMinRes
from ._types import ScipyState, ScipyStats

__all__ = [
    "ScipyBiCG",
    "ScipyBiCGStab",
    "ScipyCG",
    "ScipyMinRes",
    "ScipySolver",
    "ScipyState",
    "ScipyStats",
]



# ruff: noqa: SLF001

from __future__ import annotations

import functools
from collections.abc import Callable, Iterable, Sequence
from typing import TYPE_CHECKING, Any, Self, overload

import attrs
import equinox as eqx
from jaxtyping import Array, PyTree

from liblaf import grapes

if TYPE_CHECKING:
    from ._wrapper import FunctionWrapper


@attrs.define(kw_only=True)
class FunctionDescriptor:
    name: str | None = None
    n_outputs: int = 1
    flatten_outputs: Iterable[int] = (0,)
    unflatten_inputs: Iterable[int] = (0,)

    @overload
    def __get__(self, instance: None, owner: type) -> Self: ...
    @overload
    def __get__(
        self, instance: FunctionWrapper, owner: type | None = None
    ) -> Callable | None: ...
    def __get__(
        self, instance: FunctionWrapper | None, owner: type | None = None
    ) -> Callable | Self | None:
        assert self.name is not None
        if instance is None:
            return None
        if (cached := getattr(instance, self.wrapper_name, None)) is not None:
            return cached
        wrapped: Callable | None = getattr(instance, self.wrapped_name, None)
        if wrapped is None:
            return None

        def wrapper(
            *args: Any,
            flatten: bool = instance._flatten,
            with_aux: bool = instance._with_aux,
            **kwargs: Any,
        ) -> Any:
            __tracebackhide__ = True
            args = (*args, *instance._args)
            kwargs = {**instance._kwargs, **kwargs}
            if flatten:
                if self.name == "hess":
                    raise NotImplementedError
                assert instance.flat_def is not None
                args = _unflatten_inputs(
                    args,
                    unflatten=instance.flat_def.unflatten,
                    indices=self.unflatten_inputs,
                )
            outputs: Sequence[Any] = _as_tuple(wrapped(*args, **kwargs))
            if flatten:
                assert instance.flat_def is not None
                outputs = _flatten_outputs(
                    outputs,
                    flatten=instance.flat_def.flatten,
                    indices=self.flatten_outputs,
                )
            outputs = _with_aux(outputs, n_outputs=self.n_outputs, with_aux=with_aux)
            return outputs[0] if len(outputs) == 1 else outputs

        if instance._jit:
            wrapper = eqx.filter_jit(wrapper)
        if instance._timer:
            wrapper = grapes.timer(wrapper, label=f"{self.name}()")
        setattr(instance, self.wrapper_name, wrapper)
        return wrapper

    def __set__(self, instance: FunctionWrapper, value: Callable | None) -> None:
        setattr(instance, self.wrapped_name, value)
        setattr(instance, self.wrapper_name, None)

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    @functools.cached_property
    def wrapped_name(self) -> str:
        assert self.name is not None
        return f"_{self.name}_wrapped"

    @functools.cached_property
    def wrapper_name(self) -> str:
        assert self.name is not None
        return f"_{self.name}_wrapper"


def _as_tuple(outputs: Any) -> tuple[Any, ...]:
    if isinstance(outputs, tuple):
        return outputs
    return (outputs,)


def _flatten_outputs(
    outputs: Sequence[PyTree],
    flatten: Callable[[PyTree], Array],
    indices: Iterable[int],
) -> tuple[PyTree, ...]:
    outputs = list(outputs)
    for i in indices:
        outputs[i] = flatten(outputs[i])
    return tuple(outputs)


def _unflatten_inputs(
    inputs: Sequence[Array],
    unflatten: Callable[[Array], PyTree],
    indices: Iterable[int],
) -> tuple[PyTree, ...]:
    inputs = list(inputs)
    for i in indices:
        inputs[i] = unflatten(inputs[i])
    return tuple(inputs)


def _with_aux(
    outputs: Sequence[Any], n_outputs: int, *, with_aux: bool
) -> Sequence[Any]:
    if with_aux:
        if len(outputs) == n_outputs:
            return *outputs, None
        if len(outputs) == n_outputs + 1:
            return outputs
        raise ValueError(outputs)
    if len(outputs) == n_outputs:
        return outputs
    if len(outputs) == n_outputs + 1:
        return outputs[:-1]
    raise ValueError(outputs)



from typing import Any, override

import numpy as np
import scipy
from jaxtyping import Float

from liblaf.peach import tree
from liblaf.peach.linalg.system import LinearSystem

from ._base import ScipySolver

type FreeNp = Float[np.ndarray, " free"]


@tree.define
class ScipyMinRes(ScipySolver):
    shift: float = tree.field(default=0.0, kw_only=True)
    show: bool = tree.field(default=False, kw_only=True)
    check: bool = tree.field(default=False, kw_only=True)

    @override
    def _options(self, system: LinearSystem) -> dict[str, Any]:
        options: dict[str, Any] = super()._options(system)
        options.update({"shift": self.shift, "show": self.show, "check": self.check})
        options.pop("atol", None)
        return options

    @override
    def _wrapped(self, *args, **kwargs) -> tuple[FreeNp, int]:
        return scipy.sparse.linalg.minres(*args, **kwargs)



from __future__ import annotations

from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Self

import attrs
from jaxtyping import Array, PyTree, Shaped

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint, FixedConstraint
from liblaf.peach.tree import FlatDef


@tree.define
class FunctionWrapper:
    flat_def: FlatDef[PyTree] | None = tree.field(default=None, kw_only=True)
    _flatten: bool = tree.field(default=False, kw_only=True, alias="flatten")

    def flatten(
        self, params: PyTree, *, constraints: Iterable[Constraint] = ()
    ) -> tuple[Self, Shaped[Array, " free"], list[Constraint]]:
        fixed_constr: list[FixedConstraint] = []
        other_constr: list[Constraint] = []
        for c in constraints:
            if isinstance(c, FixedConstraint):
                fixed_constr.append(c)
            else:
                other_constr.append(c)
        if len(fixed_constr) > 1:
            raise NotImplementedError
        fixed_mask: PyTree | None = fixed_constr[0].mask if fixed_constr else None
        params_flat: Shaped[Array, " free"]
        flat_def: FlatDef[PyTree]
        params_flat, flat_def = tree.flatten(params, fixed_mask=fixed_mask)
        self_new: Self = attrs.evolve(self, flatten=True, flat_def=flat_def)
        constr_flat: list[Constraint] = [
            constr.flatten(flat_def) for constr in other_constr
        ]
        return self_new, params_flat, constr_flat

    _jit: bool = tree.field(default=False, kw_only=True, alias="jit")

    def jit(self, enable: bool = True) -> Self:  # noqa: FBT001, FBT002
        return attrs.evolve(self, jit=enable)

    _args: Sequence[Any] = tree.field(default=(), kw_only=True, alias="args")
    _kwargs: Mapping[str, Any] = tree.field(factory=dict, kw_only=True, alias="kwargs")

    def partial(self, *args: Any, **kwargs: Any) -> Self:
        return attrs.evolve(
            self, args=(*self._args, *args), kwargs={**self._kwargs, **kwargs}
        )

    _timer: bool = tree.field(default=False, kw_only=True, alias="timer")

    def timer(self, enable: bool = True) -> Self:  # noqa: FBT001, FBT002
        return attrs.evolve(self, timer=enable)

    _with_aux: bool = tree.field(default=False, kw_only=True, alias="with_aux")

    def with_aux(self, enable: bool = True) -> Self:  # noqa: FBT001, FBT002
        return attrs.evolve(self, with_aux=enable)



from typing import override

import jax
from jaxtyping import Array, Shaped

from liblaf.peach import tree

from ._base import JaxSolver

type Vector = Shaped[Array, " free"]


@tree.define
class JaxCG(JaxSolver):
    @override
    def _wrapped(self, *args, **kwargs) -> tuple[Vector, None]:
        return jax.scipy.sparse.linalg.cg(*args, **kwargs)



from __future__ import annotations

import abc
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, override

import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf import grapes
from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.linalg.abc import (
    Callback,
    LinearSolution,
    LinearSolver,
    Params,
    Result,
    SetupResult,
)
from liblaf.peach.linalg.system import LinearSystem

if TYPE_CHECKING:
    import cupy as cp
    from cupyx.scipy.sparse import linalg


type Free = Float[Array, " free"]
type FreeCp = Float[cp.ndarray, " free"]


@tree.define
class CupySolver(LinearSolver):
    from ._types import CupyState as State
    from ._types import CupyStats as Stats

    Solution = LinearSolution[State, Stats]

    rtol: float = tree.field(default=1e-5, kw_only=True)
    atol: float = tree.field(default=0.0, kw_only=True)
    max_steps: int | None = tree.field(default=None, kw_only=True)

    @override
    def setup(
        self,
        system: LinearSystem,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[State, Stats]:
        params_flat: Free
        system, params_flat, constraints = system.flatten(
            params, constraints=constraints
        )
        state: CupySolver.State = self.State(
            params_flat=params_flat, flat_def=system.flat_def
        )
        if self.jit:
            system = system.jit()
        if self.timer:
            system = system.timer()
        return SetupResult(system, constraints, state, self.Stats())

    @override
    def _solve(
        self,
        system: LinearSystem,
        state: State,
        stats: Stats,
        *,
        callback: Callback[State, Stats] | None = None,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[State, Stats, Result]:
        if constraints:
            raise NotImplementedError
        cb_wrapper: Callable = self._make_callback(callback, state, stats)
        lop: linalg.LinearOperator = _as_lop(system)
        x: FreeCp
        info: int
        x, info = self._wrapped(
            lop,
            system.b_flat,
            state.params_flat,
            callback=cb_wrapper,
            **self._options(system),
        )
        state.params_flat = jnp.from_dlpack(x)
        stats.n_steps = len(grapes.get_timer(cb_wrapper))
        result: Result
        stats, result = self._finalize(info, stats)
        return state, stats, result

    def _make_callback(
        self, callback: Callback[State, Stats] | None, state: State, stats: Stats
    ) -> Callable:
        @grapes.timer(label=f"{self.name}.callback()")
        def wrapper(xk: FreeCp) -> None:
            if callback is None:
                return
            state.params_flat = jnp.from_dlpack(xk)
            stats.n_steps = len(grapes.get_timer(wrapper)) + 1
            callback(state, stats)

        return wrapper

    def _options(self, system: LinearSystem) -> dict[str, Any]:
        options: dict[str, Any] = {"tol": self.rtol, "atol": self.atol}
        if self.max_steps is not None:
            options["maxiter"] = self.max_steps
        if system.preconditioner is not None:
            options["M"] = _preconditioner(system)
        return options

    def _finalize(self, info: int, stats: Stats) -> tuple[Stats, Result]:
        stats.info = info
        if info == 0:
            return stats, Result.SUCCESS
        if info < 0:
            return stats, Result.BREAKDOWN
        stats.n_steps = info
        return stats, Result.MAX_STEPS_REACHED

    @abc.abstractmethod
    def _wrapped(self, *args, **kwargs) -> tuple[FreeCp, int]:
        raise NotImplementedError


def _as_lop(system: LinearSystem) -> linalg.LinearOperator:
    import cupy as cp
    from cupyx.scipy.sparse import linalg

    assert system.matvec is not None

    def matvec(x: FreeCp) -> FreeCp:
        assert system.matvec is not None
        x_jax: Free = jnp.from_dlpack(x)
        y_jax: Free = system.matvec(x_jax)
        return cp.from_dlpack(y_jax)

    def rmatvec(x: FreeCp) -> FreeCp:
        assert system.rmatvec is not None
        x_jax: Free = jnp.from_dlpack(x)
        y_jax: Free = system.rmatvec(x_jax)
        return cp.from_dlpack(y_jax)

    dim: int
    (dim,) = system.b_flat.shape
    return linalg.LinearOperator(
        shape=(dim, dim),
        matvec=matvec,
        rmatvec=None if system.rmatvec is None else rmatvec,
        dtype=system.b_flat.dtype,
    )


def _preconditioner(system: LinearSystem) -> linalg.LinearOperator | None:
    import cupy as cp
    from cupyx.scipy.sparse import linalg

    if system.preconditioner is None:
        return None

    def matvec(x: FreeCp) -> FreeCp:
        assert system.preconditioner is not None
        x_jax: Free = jnp.from_dlpack(x)
        y_jax: Free = system.preconditioner(x_jax)
        return cp.from_dlpack(y_jax)

    def rmatvec(x: FreeCp) -> FreeCp:
        assert system.rpreconditioner is not None
        x_jax: Free = jnp.from_dlpack(x)
        y_jax: Free = system.rpreconditioner(x_jax)
        return cp.from_dlpack(y_jax)

    dim: int
    (dim,) = system.b_flat.shape
    return linalg.LinearOperator(
        shape=(dim, dim),
        matvec=matvec,
        rmatvec=None if system.rpreconditioner is None else rmatvec,
        dtype=system.b_flat.dtype,
    )



from ._base import JaxSolver
from ._bicgstab import JaxBiCGStab
from ._cg import JaxCG
from ._gmres import JaxGMRES
from ._types import JaxState, JaxStats

__all__ = [
    "JaxBiCGStab",
    "JaxCG",
    "JaxGMRES",
    "JaxSolver",
    "JaxState",
    "JaxStats",
]



import enum
import time
from typing import Protocol

import liblaf.grapes.rich.repr as grr
import liblaf.grapes.wadler_lindig as gwd
import wadler_lindig as wl
from jaxtyping import PyTree
from rich.repr import RichReprResult

from liblaf.peach import tree

type Params = PyTree


class Callback[StateT: State, StatsT: Stats](Protocol):
    def __call__(self, state: StateT, stats: StatsT, /) -> None: ...


class Result(enum.StrEnum):
    SUCCESS = enum.auto()
    MAX_STEPS_REACHED = enum.auto()
    UNKNOWN_ERROR = enum.auto()


@tree.define
class State:
    @property
    def params(self) -> Params:
        raise NotImplementedError


@tree.define
class Stats:
    end_time: float | None = tree.field(repr=False, default=None, kw_only=True)
    n_steps: int = tree.field(default=0, kw_only=True)
    start_time: float = tree.field(repr=False, factory=time.perf_counter, kw_only=True)

    def __pdoc__(self, **kwargs) -> wl.AbstractDoc | None:
        return gwd.pdoc_rich_repr(self, **kwargs)

    def __rich_repr__(self) -> RichReprResult:
        yield from grr.rich_repr_fieldz(self)
        yield "time", self.time

    @property
    def time(self) -> float:
        if self.end_time is None:
            return time.perf_counter() - self.start_time
        return self.end_time - self.start_time


@tree.define
class OptimizeSolution[StateT: State, StatsT: Stats]:
    result: Result
    state: StateT
    stats: StatsT

    @property
    def params(self) -> Params:
        return self.state.params

    @property
    def success(self) -> bool:
        return self.result == Result.SUCCESS



from typing import Self, override

from jaxtyping import PyTree

from liblaf.peach import tree

from ._abc import Constraint


@tree.define
class BoundConstraint(Constraint):
    lower: PyTree | None = tree.field(default=None)
    upper: PyTree | None = tree.field(default=None)
    keep_feasible: bool = tree.field(default=True, kw_only=True)

    @override
    def flatten(self, flat_def: tree.FlatDef[PyTree]) -> Self:
        lower_flat: PyTree | None = None
        upper_flat: PyTree | None = None
        if self.lower is not None:
            lower_flat = flat_def.flatten(self.lower)
        if self.upper is not None:
            upper_flat = flat_def.flatten(self.upper)
        return type(self)(lower=lower_flat, upper=upper_flat)



from ._base import CupySolver
from ._cg import CupyCG
from ._minres import CupyMinRes
from ._types import CupyState, CupyStats

__all__ = [
    "CupyCG",
    "CupyMinRes",
    "CupySolver",
    "CupyState",
    "CupyStats",
]



from __future__ import annotations

from typing import TYPE_CHECKING, override

from jaxtyping import Float

from liblaf.peach import tree

from ._base import CupySolver

if TYPE_CHECKING:
    import cupy as cp


type FreeCp = Float[cp.ndarray, " free"]


@tree.define
class CupyCG(CupySolver):
    @override
    def _wrapped(self, *args, **kwargs) -> tuple[FreeCp, int]:
        from cupyx.scipy.sparse import linalg

        return linalg.cg(*args, **kwargs)



from . import abc, cupy, jax, misc, scipy, system
from .abc import Callback, LinearSolution, LinearSolver, Result
from .cupy import CupyCG, CupyMinRes, CupySolver
from .jax import JaxBiCGStab, JaxCG, JaxGMRES, JaxSolver
from .misc import CompositeSolver
from .scipy import ScipyBiCG, ScipyBiCGStab, ScipyCG, ScipyMinRes, ScipySolver
from .system import LinearSystem

__all__ = [
    "Callback",
    "CompositeSolver",
    "CupyCG",
    "CupyMinRes",
    "CupySolver",
    "JaxBiCGStab",
    "JaxCG",
    "JaxGMRES",
    "JaxSolver",
    "LinearSolution",
    "LinearSolver",
    "LinearSystem",
    "Result",
    "ScipyBiCG",
    "ScipyBiCGStab",
    "ScipyCG",
    "ScipyMinRes",
    "ScipySolver",
    "abc",
    "cupy",
    "jax",
    "misc",
    "scipy",
    "system",
]



from typing import Any, Literal, override

import jax
from jaxtyping import Array, ArrayLike, Integer, Shaped

from liblaf.peach import tree
from liblaf.peach.linalg.system import LinearSystem

from ._base import JaxSolver

type Vector = Shaped[Array, " free"]


@tree.define
class JaxGMRES(JaxSolver):
    restart: int = 20
    solve_method: Literal["incremental", "batched"] = "batched"

    @override
    def _options(self, system: LinearSystem) -> dict[str, Any]:
        options: dict[str, Any] = super()._options(system)
        options.update({"restart": self.restart, "solve_method": self.solve_method})
        return options

    @override
    def _wrapped(self, *args, **kwargs) -> tuple[Vector, Integer[ArrayLike, ""]]:
        return jax.scipy.sparse.linalg.gmres(*args, **kwargs)



# ruff: noqa: N803, N806

from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, override

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Array, Float, ScalarLike

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.optim.abc import (
    Optimizer,
    OptimizeSolution,
    Params,
    Result,
    SetupResult,
)
from liblaf.peach.optim.objective import Objective

from ._state import PNCGState
from ._stats import PNCGStats

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]


@tree.define
class PNCG(Optimizer[PNCGState, PNCGStats]):
    from ._state import PNCGState as State
    from ._stats import PNCGStats as Stats

    Solution = OptimizeSolution[PNCGState, PNCGStats]

    max_steps: int = tree.field(default=256, kw_only=True)
    norm: Callable[[Params], Scalar] | None = tree.field(default=None, kw_only=True)

    if TYPE_CHECKING:
        atol: ScalarLike = tree.field(default=1e-28, kw_only=True)
        rtol: ScalarLike = tree.field(default=1e-5, kw_only=True)
        d_hat: ScalarLike = tree.field(default=jnp.inf, kw_only=True)
    else:
        atol: Scalar = tree.array(default=1e-28, kw_only=True)
        rtol: Scalar = tree.array(default=1e-5, kw_only=True)
        d_hat: Scalar = tree.array(default=jnp.inf, kw_only=True)

    @override
    def init(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[PNCGState, PNCGStats]:
        params_flat: Vector
        objective, params_flat, constraints = objective.flatten(
            params, constraints=constraints
        )
        if self.jit:
            objective = objective.jit()
        if self.timer:
            objective = objective.timer()
        state = PNCGState(params_flat=params_flat, flat_def=objective.flat_def)
        return SetupResult(objective, constraints, state, PNCGStats())

    @override
    def step(
        self,
        objective: Objective,
        state: PNCGState,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> PNCGState:
        if constraints:
            raise NotImplementedError
        assert objective.grad_and_hess_diag is not None
        assert objective.hess_quad is not None
        g: Vector
        H_diag: Vector
        g, H_diag = objective.grad_and_hess_diag(state.params_flat)
        P: Vector = jnp.reciprocal(H_diag)
        beta: Scalar
        p: Vector
        if state.search_direction_flat is None:
            beta = jnp.zeros(())
            p = -P * g
        else:
            beta = self._compute_beta(
                g_prev=state.grad_flat, g=g, p=state.search_direction_flat, P=P
            )
            p = -P * g + beta * state.search_direction_flat
        pHp: Scalar = objective.hess_quad(state.params_flat, p)
        alpha: Scalar = self._compute_alpha(g=g, p=p, pHp=pHp, unflatten=state.flat_def)
        state.params_flat += alpha * p
        DeltaE: Scalar = -alpha * jnp.vdot(g, p) - 0.5 * alpha**2 * pHp
        if state.first_decrease is None:
            state.first_decrease = DeltaE
        state.alpha = alpha
        state.beta = beta
        state.decrease = DeltaE
        state.grad_flat = g
        state.hess_diag_flat = H_diag
        state.hess_quad = pHp
        state.preconditioner_flat = P
        state.search_direction_flat = p
        return state

    @override
    def terminate(
        self,
        objective: Objective,
        state: PNCGState,
        stats: PNCGStats,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[bool, Result]:
        if constraints:
            raise NotImplementedError
        assert state.first_decrease is not None
        if state.decrease < self.atol + self.rtol * state.first_decrease:
            return True, Result.SUCCESS
        if stats.n_steps >= self.max_steps:
            return True, Result.MAX_STEPS_REACHED
        return False, Result.UNKNOWN_ERROR

    @eqx.filter_jit
    def _compute_alpha(
        self,
        g: Vector,
        p: Vector,
        pHp: Scalar,
        unflatten: Callable[[Array], Params] | None = None,
    ) -> Scalar:
        p_norm: Scalar
        if self.norm is None:
            p_norm = jnp.linalg.norm(p, ord=jnp.inf)
        else:
            p_tree: Params = p if unflatten is None else unflatten(p)
            p_norm = self.norm(p_tree)
        alpha_1: Scalar = self.d_hat / (2.0 * p_norm)  # pyright: ignore[reportAssignmentType]
        alpha_2: Scalar = -jnp.vdot(g, p) / pHp
        alpha: Scalar = jnp.minimum(alpha_1, alpha_2)
        alpha = jnp.nan_to_num(alpha)
        return alpha

    @eqx.filter_jit
    def _compute_beta(self, g_prev: Vector, g: Vector, p: Vector, P: Vector) -> Scalar:
        y: Vector = g - g_prev
        yTp: Scalar = jnp.vdot(y, p)
        Py: Scalar = P * y
        beta: Scalar = jnp.vdot(g, Py) / yTp - (jnp.vdot(y, Py) / yTp) * (
            jnp.vdot(p, g) / yTp
        )
        return beta



import abc
import time
from collections.abc import Iterable
from typing import NamedTuple

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.optim.objective import Objective

from ._types import Callback, OptimizeSolution, Params, Result, State, Stats


class SetupResult[StateT: State, StatsT: Stats](NamedTuple):
    objective: Objective
    constraints: list[Constraint]
    state: StateT
    stats: StatsT


@tree.define
class Optimizer[StateT: State, StatsT: Stats](abc.ABC):
    from ._types import OptimizeSolution as Solution
    from ._types import State, Stats

    max_steps: int = tree.field(default=256, kw_only=True)
    jit: bool = tree.field(default=False, kw_only=True)
    timer: bool = tree.field(default=False, kw_only=True)

    @abc.abstractmethod
    def init(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[StateT, StatsT]:
        raise NotImplementedError

    @abc.abstractmethod
    def step(
        self,
        objective: Objective,
        state: StateT,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> StateT:
        raise NotImplementedError

    def update_stats(
        self,
        objective: Objective,  # noqa: ARG002
        state: StateT,  # noqa: ARG002
        stats: StatsT,
        *,
        constraints: Iterable[Constraint] = (),  # noqa: ARG002
    ) -> StatsT:
        return stats

    @abc.abstractmethod
    def terminate(
        self,
        objective: Objective,
        state: StateT,
        stats: StatsT,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[bool, Result]:
        raise NotImplementedError

    def postprocess(
        self,
        objective: Objective,  # noqa: ARG002
        state: StateT,
        stats: StatsT,
        result: Result,
        *,
        constraints: Iterable[Constraint] = (),  # noqa: ARG002
    ) -> OptimizeSolution[StateT, StatsT]:
        stats.end_time = time.perf_counter()
        solution: OptimizeSolution[StateT, StatsT] = OptimizeSolution(
            result=result, state=state, stats=stats
        )
        return solution

    def minimize(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
        callback: Callback[StateT, StatsT] | None = None,
    ) -> OptimizeSolution[StateT, StatsT]:
        state: StateT
        stats: StatsT
        objective, constraints, state, stats = self.init(
            objective, params, constraints=constraints
        )
        done: bool = False
        n_steps: int = 0
        result: Result = Result.UNKNOWN_ERROR
        while n_steps < self.max_steps and not done:
            state = self.step(objective, state, constraints=constraints)
            n_steps += 1
            stats.n_steps = n_steps
            stats = self.update_stats(objective, state, stats, constraints=constraints)
            if callback is not None:
                callback(state, stats)
            done, result = self.terminate(
                objective, state, stats, constraints=constraints
            )
        if not done:
            result = Result.MAX_STEPS_REACHED
        solution: OptimizeSolution[StateT, StatsT] = self.postprocess(
            objective, state, stats, result, constraints=constraints
        )
        return solution



from collections.abc import Iterator, Mapping
from typing import Any

from scipy.optimize import OptimizeResult

from liblaf.peach import tree
from liblaf.peach.optim.abc import Params, State
from liblaf.peach.tree import FlatDef


@tree.define
class ScipyState(Mapping[str, Any], State):
    flat_def: FlatDef[Params]
    result: OptimizeResult = tree.container(factory=OptimizeResult)

    def __getitem__(self, key: str, /) -> Any:
        return self.result[key]

    def __iter__(self) -> Iterator[str]:
        return iter(self.result)

    def __len__(self) -> int:
        return len(self.result)

    @property
    def fun(self) -> float:
        return self.result["fun"]

    @property
    def params(self) -> Params:
        return self.flat_def.unflatten(self.result["x"])

    @params.setter
    def params(self, value: Params, /) -> None:
        self.result["x"] = self.flat_def.flatten(value)



import jax.numpy as jnp
from jaxtyping import Array, Float, PyTree

from liblaf.peach import tree
from liblaf.peach.optim.abc import State
from liblaf.peach.tree import FlatDef, TreeView

type Scalar = Float[Array, ""]
type Vector = Float[Array, " N"]
type Params = PyTree


@tree.define
class PNCGState(State):
    flat_def: FlatDef[Params] | None = tree.field(default=None, kw_only=True)

    alpha: Scalar = tree.array(default=None)
    """line search step size"""

    beta: Scalar = tree.array(default=jnp.zeros(()))
    """Dai-Kou (DK) algorithm"""

    decrease: Scalar = tree.array(default=None)
    """Delta E"""

    first_decrease: Scalar = tree.array(default=None)
    """Delta E_0"""

    grad = TreeView[Params]()
    """g"""
    grad_flat: Vector = tree.array(default=None)

    hess_diag = TreeView[Params]()
    """diag(H)"""
    hess_diag_flat: Vector = tree.array(default=None)

    hess_quad: Scalar = tree.array(default=None)
    """pHp"""

    params = TreeView[Params]()  # pyright: ignore[reportIncompatibleMethodOverride, reportAssignmentType]
    """x"""
    params_flat: Vector = tree.array(default=None)

    preconditioner = TreeView[Params]()
    """P"""
    preconditioner_flat: Vector = tree.array(default=None)

    search_direction = TreeView[Params]()
    """p"""
    search_direction_flat: Vector = tree.array(default=None)



from ._define import define
from ._field import array, container, field, static
from ._flatten import FlatDef, flatten
from ._register_fieldz import register_fieldz
from ._view import FlatView, TreeView

__all__ = [
    "FlatDef",
    "FlatView",
    "TreeView",
    "array",
    "container",
    "define",
    "field",
    "flatten",
    "register_fieldz",
    "static",
]



import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from liblaf.peach import math
from liblaf.peach.optim import Objective

type Vector = Float[Array, " N"]
type Scalar = Float[Array, ""]


@jax.jit
def rosen(x: Vector) -> Scalar:
    return jnp.sum(
        100.0 * jnp.square(x[1:] - jnp.square(x[:-1])) + jnp.square(1.0 - x[:-1])
    )


@jax.jit
def rosen_grad(x: Vector) -> Vector:
    return jax.grad(rosen)(x)


@jax.jit
def rosen_hess_diag(x: Vector) -> Vector:
    return jnp.diagonal(jax.hessian(rosen)(x))


@jax.jit
def rosen_hess_prod(x: Vector, p: Vector) -> Vector:
    return math.hess_prod(rosen, x, p)


@jax.jit
def rosen_hess_quad(x: Vector, p: Vector) -> Scalar:
    return jnp.vdot(p, rosen_hess_prod(x, p))


@jax.jit
def rosen_value_and_grad(x: Vector) -> tuple[Scalar, Vector]:
    return jax.value_and_grad(rosen)(x)


@jax.jit
def rosen_grad_and_hess_diag(x: Vector) -> tuple[Vector, Vector]:
    return rosen_grad(x), rosen_hess_diag(x)


def rosen_objective() -> Objective:
    return Objective(
        fun=rosen,
        grad=rosen_grad,
        hess_diag=rosen_hess_diag,
        hess_prod=rosen_hess_prod,
        hess_quad=rosen_hess_quad,
        value_and_grad=rosen_value_and_grad,
        grad_and_hess_diag=rosen_grad_and_hess_diag,
    )



from __future__ import annotations

from collections.abc import Callable, Iterable, Mapping
from typing import TYPE_CHECKING, Any, Never, override

import numpy as np
import scipy
from jaxtyping import Array
from scipy.optimize import Bounds, OptimizeResult

from liblaf import grapes
from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.constraints._bound import BoundConstraint
from liblaf.peach.optim.abc import (
    Callback,
    Optimizer,
    OptimizeSolution,
    Params,
    Result,
    SetupResult,
)
from liblaf.peach.optim.objective import Objective
from liblaf.peach.tree import FlatDef

from ._state import ScipyState
from ._stats import ScipyStats

if TYPE_CHECKING:
    from scipy.optimize._minimize import _CallbackResult


@tree.define
class ScipyOptimizer(Optimizer[ScipyState, ScipyStats]):
    from ._state import ScipyState as State
    from ._stats import ScipyStats as Stats

    Solution = OptimizeSolution[ScipyState, ScipyStats]

    method: str | None = tree.field(default=None, kw_only=True)
    tol: float | None = tree.field(default=None, kw_only=True)
    options: Mapping[str, Any] | None = tree.field(default=None, kw_only=True)

    @override
    def init(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[ScipyState, ScipyStats]:
        params_flat: Array
        objective, params_flat, constraints = objective.flatten(
            params, constraints=constraints
        )
        if self.jit:
            objective = objective.jit()
        if self.timer:
            objective = objective.timer()
        assert objective.flat_def is not None
        state = ScipyState(
            flat_def=objective.flat_def, result=OptimizeResult({"x": params_flat})
        )
        stats = ScipyStats()
        return SetupResult(objective, constraints, state, stats)

    @override
    def step(
        self,
        objective: Objective,
        state: ScipyState,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> Never:
        raise NotImplementedError

    @override
    def terminate(
        self,
        objective: Objective,
        state: ScipyState,
        stats: ScipyStats,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> Never:
        raise NotImplementedError

    @override
    def minimize(
        self,
        objective: Objective,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
        callback: Callback[ScipyState, ScipyStats] | None = None,
    ) -> OptimizeSolution[ScipyState, ScipyStats]:
        options: dict[str, Any] = {"maxiter": self.max_steps}
        if self.options is not None:
            options.update(self.options)
        state: ScipyState
        stats: ScipyStats
        objective, constraints, state, stats = self.init(
            objective, params, constraints=constraints
        )
        callback_wrapper: _CallbackResult = self._make_callback(
            objective, callback, stats, state.flat_def
        )
        fun: Callable | None
        jac: Callable | bool | None
        if objective.value_and_grad is None:
            fun = objective.fun
            jac = objective.grad
        else:
            fun = objective.value_and_grad
            jac = True
        raw: OptimizeResult = scipy.optimize.minimize(  # pyright: ignore[reportCallIssue]
            bounds=self._make_bounds(constraints),
            callback=callback_wrapper,
            fun=fun,  # pyright: ignore[reportArgumentType]
            hess=objective.hess,
            hessp=objective.hess_prod,
            jac=jac,  # pyright: ignore[reportArgumentType]
            method=self.method,  # pyright: ignore[reportArgumentType]
            options=options,  # pyright: ignore[reportArgumentType]
            tol=self.tol,
            x0=state.result["x"],
        )
        state: ScipyState = self._unflatten_state(raw, state.flat_def)
        result: Result = (
            Result.SUCCESS if state.result["success"] else Result.UNKNOWN_ERROR
        )
        solution: OptimizeSolution[ScipyState, ScipyStats] = self.postprocess(
            objective, state, stats, result
        )
        return solution

    def _make_bounds(self, constraints: list[Constraint]) -> Bounds | None:
        bound_constr: list[BoundConstraint] = []
        other_constr: list[Constraint] = []
        for c in constraints:
            if isinstance(c, BoundConstraint):
                bound_constr.append(c)
            else:
                other_constr.append(c)
        if other_constr:
            raise NotImplementedError
        if not bound_constr:
            return None
        if len(bound_constr) > 1:
            raise NotImplementedError
        bound: BoundConstraint = bound_constr[0]
        return Bounds(
            -np.inf if bound.lower is None else bound.lower,
            np.inf if bound.upper is None else bound.upper,
        )

    def _make_callback(
        self,
        objective: Objective,
        callback: Callback[ScipyState, ScipyStats] | None,
        stats: ScipyStats,
        unflatten: FlatDef[Params],
    ) -> _CallbackResult:
        @grapes.timer(label="callback()")
        def wrapper(intermediate_result: OptimizeResult) -> None:
            nonlocal stats
            if callback is not None:
                state: ScipyState = self._unflatten_state(
                    intermediate_result, unflatten
                )
                stats.n_steps = len(grapes.get_timer(wrapper)) + 1
                stats = self.update_stats(objective, state, stats)
                callback(state, stats)

        return wrapper

    def _unflatten_state(
        self, result: OptimizeResult, unflatten: FlatDef[Params]
    ) -> ScipyState:
        state = ScipyState(result=result, flat_def=unflatten)
        return state



from . import constraints, cuda, functools, linalg, optim, tree
from ._version import __version__, __version_tuple__
from .constraints import Constraint, FixedConstraint

__all__ = [
    "Constraint",
    "FixedConstraint",
    "__version__",
    "__version_tuple__",
    "constraints",
    "cuda",
    "functools",
    "linalg",
    "optim",
    "tree",
]



from typing import override

import numpy as np
import scipy
from jaxtyping import Float

from liblaf.peach import tree

from ._base import ScipySolver

type FreeNp = Float[np.ndarray, " free"]


@tree.define
class ScipyCG(ScipySolver):
    @override
    def _wrapped(self, *args, **kwargs) -> tuple[FreeNp, int]:
        return scipy.sparse.linalg.cg(*args, **kwargs)



from collections.abc import Generator, Iterable, Sequence
from typing import Any, Self

import attrs
import fieldz
import jax.tree_util as jtu

type AuxData = Sequence[Any]
type Children = Iterable[Any]
type ChildrenWithKeys = Iterable[tuple[Any, Any]]


@attrs.define
class Flattener[T]:
    cls: type[T]
    data_fields: Iterable[str] = ()
    meta_fields: Iterable[str] = ()

    @classmethod
    def from_cls(
        cls,
        nodetype: type[T],
        data_fields: Iterable[str] | None = None,
        meta_fields: Iterable[str] | None = None,
    ) -> Self:
        if data_fields is None:
            data_fields = _filter_fields(nodetype, static=False)
        if meta_fields is None:
            meta_fields = _filter_fields(nodetype, static=True)
        return cls(
            cls=nodetype, data_fields=tuple(data_fields), meta_fields=tuple(meta_fields)
        )

    def flatten_with_keys(self, obj: Any) -> tuple[ChildrenWithKeys, AuxData]:
        children: ChildrenWithKeys = tuple(
            (jtu.GetAttrKey(name), getattr(obj, name)) for name in self.data_fields
        )
        aux_data: AuxData = tuple(getattr(obj, name) for name in self.meta_fields)
        return children, aux_data

    def flatten(self, obj: Any) -> tuple[Children, AuxData]:
        children: Children = tuple(getattr(obj, name) for name in self.data_fields)
        aux_data: AuxData = tuple(getattr(obj, name) for name in self.meta_fields)
        return children, aux_data

    def unflatten(self, aux_data: AuxData, children: Children) -> T:
        obj: T = object.__new__(self.cls)
        for name, value in zip(self.meta_fields, aux_data, strict=True):
            object.__setattr__(obj, name, value)
        for name, value in zip(self.data_fields, children, strict=True):
            object.__setattr__(obj, name, value)
        return obj


def register_fieldz[T: type](
    cls: T,
    data_fields: Iterable[str] | None = None,
    meta_fields: Iterable[str] | None = None,
) -> T:
    flattener: Flattener[T] = Flattener.from_cls(
        cls, data_fields=data_fields, meta_fields=meta_fields
    )
    jtu.register_pytree_with_keys(
        cls,
        flatten_with_keys=flattener.flatten_with_keys,
        unflatten_func=flattener.unflatten,
        flatten_func=flattener.flatten,
    )
    return cls


def _filter_fields(cls: type, *, static: bool) -> Generator[str]:
    for field in fieldz.fields(cls):
        if field.metadata.get("static", False) == static:
            yield field.name



from __future__ import annotations

from typing import TYPE_CHECKING, Any, override

from jaxtyping import Float

from liblaf.peach import tree
from liblaf.peach.linalg.system import LinearSystem

from ._base import CupySolver

if TYPE_CHECKING:
    import cupy as cp


type FreeCp = Float[cp.ndarray, " free"]


@tree.define
class CupyMinRes(CupySolver):
    shift: float = tree.field(default=0.0, kw_only=True)

    @override
    def _options(self, system: LinearSystem) -> dict[str, Any]:
        options: dict[str, Any] = super()._options(system)
        options.update({"shift": self.shift})
        options.pop("atol", None)
        return options

    @override
    def _wrapped(self, *args, **kwargs) -> tuple[FreeCp, int]:
        from cupyx.scipy.sparse import linalg

        return linalg.minres(*args, **kwargs)



import abc
import functools
import time
from collections.abc import Iterable
from typing import NamedTuple

from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.linalg.system import LinearSystem

from ._types import Callback, LinearSolution, Params, Result, State, Stats

type Vector = Float[Array, " free"]


class SetupResult[StateT: State, StatsT: Stats](NamedTuple):
    system: LinearSystem
    constraints: list[Constraint]
    state: StateT
    stats: StatsT


@tree.define
class LinearSolver[StateT: State, StatsT: Stats](abc.ABC):
    from ._types import LinearSolution as Solution
    from ._types import State, Stats

    jit: bool = tree.field(default=False, kw_only=True)
    timer: bool = tree.field(default=False, kw_only=True)

    @functools.cached_property
    def name(self) -> str:
        return type(self).__name__

    def setup(
        self,
        system: LinearSystem,
        params: Params,
        *,
        constraints: Iterable[Constraint] = (),
    ) -> SetupResult[StateT, StatsT]:
        params_flat: Vector
        system, params_flat, constraints = system.flatten(
            params, constraints=constraints
        )
        state = self.State(params_flat=params_flat, flat_def=system.flat_def)
        if self.jit:
            system = system.jit()
        if self.timer:
            system = system.timer()
        return SetupResult(system, constraints, state, self.Stats())  # pyright: ignore[reportReturnType]

    def finalize(
        self,
        system: LinearSystem,  # noqa: ARG002
        state: StateT,
        stats: StatsT,
        result: Result,
    ) -> LinearSolution[StateT, StatsT]:
        stats.end_time = time.perf_counter()
        return LinearSolution(state=state, stats=stats, result=result)

    def solve(
        self,
        system: LinearSystem,
        params: Params,
        *,
        callback: Callback[StateT, StatsT] | None = None,
        constraints: Iterable[Constraint] = (),
    ) -> LinearSolution[StateT, StatsT]:
        state: StateT
        stats: StatsT
        system, constraints, state, stats = self.setup(
            system, params, constraints=constraints
        )
        result: Result
        state, stats, result = self._solve(
            system, state, stats, callback=callback, constraints=constraints
        )
        return self.finalize(system, state, stats, result)

    @abc.abstractmethod
    def _solve(
        self,
        system: LinearSystem,
        state: StateT,
        stats: StatsT,
        *,
        callback: Callback[StateT, StatsT] | None = None,
        constraints: Iterable[Constraint] = (),
    ) -> tuple[StateT, StatsT, Result]:
        raise NotImplementedError



from collections.abc import Callable, Iterable

import equinox as eqx
import jax.flatten_util as jfu
import jax.numpy as jnp
from jaxtyping import Array, Bool, DTypeLike, Integer, PyTree, Shaped

from ._define import define
from ._field import static


@define
class FlatDef[T]:
    full_flat: Shaped[Array, " full"]
    static: T = static()
    unravel: Callable[[Shaped[Array, " full"]], T]
    free_indices: Integer[Array, " free"] | None = None

    @eqx.filter_jit
    def flatten(self, tree: T) -> Shaped[Array, " free"]:
        data: T = eqx.filter(tree, eqx.is_array)
        full_flat: Shaped[Array, " full"]
        full_flat, _ = jfu.ravel_pytree(data)
        free_flat: Shaped[Array, " free"] = (
            full_flat if self.free_indices is None else full_flat[self.free_indices]
        )
        return free_flat

    @eqx.filter_jit
    def unflatten(
        self, free_flat: Shaped[Array, " free"], dtype: DTypeLike | None = None
    ) -> T:
        if dtype is None:
            dtype = self.full_flat.dtype
        free_flat = jnp.asarray(free_flat, dtype)
        full_flat: Shaped[Array, " data"] = (
            free_flat
            if self.free_indices is None
            else self.full_flat.at[self.free_indices].set(free_flat)
        )
        data: T = self.unravel(full_flat)
        tree: T = eqx.combine(data, self.static)
        return tree


def flatten[T](
    obj: T,
    *,
    fixed_mask: T | None = None,
    fixed_selector: Callable[[T], Iterable[Array]] | None = None,
) -> tuple[Array, FlatDef[T]]:
    # TODO: JIT?
    return _flatten(obj, fixed_mask=fixed_mask, fixed_selector=fixed_selector)


def _flatten[T](
    obj: T,
    *,
    fixed_mask: T | None = None,
    fixed_selector: Callable[[T], Iterable[Array]] | None = None,
) -> tuple[Array, FlatDef[T]]:
    data: T
    static: T
    data, static = eqx.partition(obj, eqx.is_array)
    full_flat: Shaped[Array, " full"]
    unravel: Callable[[Array], T]
    full_flat, unravel = jfu.ravel_pytree(data)
    free_indices: Integer[Array, " free"] | None = None
    if fixed_mask is not None:
        free_indices = _fixed_mask_to_free_indices(fixed_mask)
    elif fixed_selector is not None:
        free_indices = _fixed_selector_to_free_indices(
            fixed_selector, n_full=full_flat.size, unravel=unravel
        )
    free_flat: Shaped[Array, " free"] = (
        full_flat[free_indices] if free_indices is not None else full_flat
    )
    return free_flat, FlatDef(
        full_flat=full_flat, unravel=unravel, static=static, free_indices=free_indices
    )


def _fixed_mask_to_free_indices(
    fixed_mask: PyTree, *, n_free: int | None = None
) -> Integer[Array, " free"]:
    fixed_mask = eqx.filter(fixed_mask, eqx.is_array)
    fixed_mask_flat: Bool[Array, " full"]
    fixed_mask_flat, _ = jfu.ravel_pytree(fixed_mask)
    free_mask_flat: Bool[Array, " full"] = ~fixed_mask_flat
    free_idx_flat: Integer[Array, " free"] = jnp.flatnonzero(
        free_mask_flat, size=n_free
    )
    return free_idx_flat


def _fixed_selector_to_free_indices[T](
    fixed_selector: Callable[[T], PyTree],
    n_full: int,
    unravel: Callable[[Array], T],
    *,
    n_free: int | None = None,
) -> Integer[Array, " free"]:
    full_idx_flat: Integer[Array, " full"] = jnp.arange(n_full)
    full_idx_tree: T = unravel(full_idx_flat)
    fixed_idx_tree: PyTree = fixed_selector(full_idx_tree)
    fixed_idx_flat: Integer[Array, " fixed"]
    fixed_idx_flat, _ = jfu.ravel_pytree(fixed_idx_tree)
    free_mask_flat: Bool[Array, " full"] = jnp.ones(n_full, jnp.bool)
    free_mask_flat = free_mask_flat.at[fixed_idx_flat].set(False)
    free_idx_flat: Integer[Array, " free"] = jnp.flatnonzero(
        free_mask_flat, size=n_free
    )
    return free_idx_flat


_flatten_jit = eqx.filter_jit(_flatten)



from __future__ import annotations

from collections.abc import Callable

from liblaf.peach import tree
from liblaf.peach.functools import FunctionDescriptor, FunctionWrapper


@tree.define
class Objective(FunctionWrapper):
    fun = FunctionDescriptor(n_outputs=1, unflatten_inputs=(0,), flatten_outputs=())
    """X -> Scalar"""
    _fun_wrapped: Callable | None = tree.field(default=None, alias="fun")
    _fun_wrapper: Callable | None = tree.field(default=None, init=False)

    grad = FunctionDescriptor(n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,))
    """X -> X"""
    _grad_wrapped: Callable | None = tree.field(default=None, alias="grad")
    _grad_wrapper: Callable | None = tree.field(default=None, init=False)

    hess = FunctionDescriptor(n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,))
    """X -> H"""
    _hess_wrapped: Callable | None = tree.field(default=None, alias="hess")
    _hess_wrapper: Callable | None = tree.field(default=None, init=False)

    hess_diag = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,)
    )
    """X -> X"""
    _hess_diag_wrapped: Callable | None = tree.field(default=None, alias="hess_diag")
    _hess_diag_wrapper: Callable | None = tree.field(default=None, init=False)

    hess_prod = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0, 1), flatten_outputs=(0,)
    )
    """X, P -> X"""
    _hess_prod_wrapped: Callable | None = tree.field(default=None, alias="hess_prod")
    _hess_prod_wrapper: Callable | None = tree.field(default=None, init=False)

    hess_quad = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0, 1), flatten_outputs=()
    )
    """X, P -> Scalar"""
    _hess_quad_wrapped: Callable | None = tree.field(default=None, alias="hess_quad")
    _hess_quad_wrapper: Callable | None = tree.field(default=None, init=False)

    value_and_grad = FunctionDescriptor(
        n_outputs=2, unflatten_inputs=(0,), flatten_outputs=(1,)
    )
    """X -> Scalar, X"""
    _value_and_grad_wrapped: Callable | None = tree.field(
        default=None, alias="value_and_grad"
    )
    _value_and_grad_wrapper: Callable | None = tree.field(default=None, init=False)

    grad_and_hess_diag = FunctionDescriptor(
        n_outputs=2, unflatten_inputs=(0,), flatten_outputs=(0, 1)
    )
    """X -> X, X"""
    _grad_and_hess_diag_wrapped: Callable | None = tree.field(
        default=None, alias="grad_and_hess_diag"
    )
    _grad_and_hess_diag_wrapper: Callable | None = tree.field(default=None, init=False)



import functools

from jaxtyping import Array

from ._flatten import FlatDef, flatten


class TreeView[T]:
    name: str
    flat_def_name: str

    def __init__(self, flat: str | None = None, flat_def: str = "flat_def") -> None:
        if flat is not None:
            self.flat_name = flat
        self.flat_def_name = flat_def

    def __get__(self, instance: object, owner: type) -> T:
        value: Array = getattr(instance, self.flat_name)
        flat_def: FlatDef[T] = getattr(instance, self.flat_def_name)
        return flat_def.unflatten(value)

    def __set__(self, instance: object, tree: T) -> None:
        flat_def: FlatDef[T] | None = getattr(instance, self.flat_def_name, None)
        flat: Array
        if flat_def is None:
            flat, flat_def = flatten(tree)
            setattr(instance, self.flat_def_name, flat_def)
        else:
            flat = flat_def.flatten(tree)
        setattr(instance, self.flat_name, flat)

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    @functools.cached_property
    def flat_name(self) -> str:
        if self.name.endswith("_tree"):
            return self.name.removesuffix("_tree")
        return f"{self.name}_flat"


class FlatView[T]:
    name: str
    flat_def_name: str

    def __init__(self, tree: str | None = None, flat_def: str = "flat_def") -> None:
        if tree is not None:
            self.tree_name = tree
        self.flat_def_name = flat_def

    def __get__(self, instance: object, owner: type) -> Array:
        tree: T = getattr(instance, self.tree_name)
        flat_def: FlatDef[T] | None = getattr(instance, self.flat_def_name, None)
        flat: Array
        if flat_def is None:
            flat, flat_def = flatten(tree)
            setattr(instance, self.flat_def_name, flat_def)
        else:
            flat = flat_def.flatten(tree)
        return flat

    def __set__(self, instance: object, flat: Array) -> None:
        flat_def: FlatDef[T] = getattr(instance, self.flat_def_name)
        tree: T = flat_def.unflatten(flat)
        setattr(instance, self.tree_name, tree)

    def __set_name__(self, owner: type, name: str) -> None:
        self.name = name

    @functools.cached_property
    def tree_name(self) -> str:
        if self.name.endswith("_flat"):
            return self.name.removesuffix("_flat")
        return f"{self.name}_tree"



from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.linalg.abc import State, Stats

type Scalar = Float[Array, ""]


@tree.define
class CupyState(State):
    pass


@tree.define
class CupyStats(Stats):
    info: int = -1
    n_steps: int | None = None



from typing import Any, overload

import attrs
import jax.numpy as jnp
import toolz
from jaxtyping import Array, ArrayLike
from liblaf.grapes import wraps


@wraps(attrs.field)
def array(**kwargs) -> Any:
    kwargs.setdefault("converter", _optional_as_array)
    return field(**kwargs)


@wraps(attrs.field)
def container(**kwargs) -> Any:
    if "converter" in kwargs and "factory" not in kwargs:
        kwargs["factory"] = kwargs["converter"]  # pyright: ignore[reportGeneralTypeIssues]
    elif "converter" not in kwargs and "factory" in kwargs:
        kwargs["converter"] = kwargs["factory"]  # pyright: ignore[reportGeneralTypeIssues]
    elif "converter" not in kwargs and "factory" not in kwargs:
        kwargs["converter"] = _dict_if_none
        kwargs["factory"] = dict
    return field(**kwargs)


@wraps(attrs.field)
def field(**kwargs) -> Any:
    if "default_factory" in kwargs:
        kwargs.setdefault("factory", kwargs.pop("default_factory"))
    if kwargs.pop("static", False):
        kwargs["metadata"] = toolz.assoc(kwargs.get("metadata") or {}, "static", True)  # noqa: FBT003
    return attrs.field(**kwargs)  # pyright: ignore[reportCallIssue]


@wraps(attrs.field)
def static(**kwargs) -> Any:
    kwargs.setdefault("static", True)
    return field(**kwargs)  # pyright: ignore[reportCallIssue]


@overload
def _dict_if_none(value: None) -> dict: ...
@overload
def _dict_if_none[T](value: T) -> T: ...
def _dict_if_none(value: Any) -> Any:
    if value is None:
        return {}
    return value


@overload
def _optional_as_array(value: None) -> None: ...
@overload
def _optional_as_array(value: ArrayLike) -> Array: ...
def _optional_as_array(value: Any) -> Any:
    if value is None:
        return None
    return jnp.asarray(value)



from ._optimizer import Optimizer, SetupResult
from ._types import Callback, OptimizeSolution, Params, Result, State, Stats

__all__ = [
    "Callback",
    "OptimizeSolution",
    "Optimizer",
    "Params",
    "Result",
    "SetupResult",
    "State",
    "Stats",
]



from collections.abc import Callable, Mapping, Sequence
from typing import Any

import equinox as eqx
from jaxtyping import PyTree


def hess_prod(
    func: Callable,
    x: PyTree,
    p: PyTree,
    args: Sequence[Any] = (),
    kwargs: Mapping[str, Any] = {},
) -> PyTree:
    def wrapper(x: PyTree) -> PyTree:
        return func(x, *args, **kwargs)

    output: PyTree
    _, output = eqx.filter_jvp(eqx.filter_grad(wrapper), (x,), (p,))
    return output



from collections.abc import Iterable
from typing import override

from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.constraints import Constraint
from liblaf.peach.linalg.abc import Callback, LinearSolution, LinearSolver, Result
from liblaf.peach.linalg.system import LinearSystem

type Vector = Float[Array, " free"]


def _default_solvers() -> list[LinearSolver]:
    from liblaf.peach.linalg.cupy import CupyCG, CupyMinRes

    return [CupyCG(), CupyMinRes()]


@tree.define
class CompositeSolver(LinearSolver):
    from ._composite_types import CompositeState as State
    from ._composite_types import CompositeStats as Stats

    Solution = LinearSolution[State, Stats]

    solvers: list[LinearSolver] = tree.field(factory=_default_solvers)

    @override
    def _solve(
        self,
        system: LinearSystem,
        state: State,
        stats: Stats,
        *,
        constraints: Iterable[Constraint] = (),
        callback: Callback[State, Stats] | None = None,
    ) -> tuple[State, Stats, Result]:
        solution: LinearSolution = None  # pyright: ignore[reportAssignmentType]
        for solver in self.solvers:
            solution = solver.solve(
                system, state.params, callback=callback, constraints=constraints
            )
            state.state.append(solution.state)
            stats.stats.append(solution.stats)
            if solution.success:
                break
        state.params_flat = solution.params_flat
        return state, stats, solution.result



from jaxtyping import PyTree

from liblaf.peach import tree

from ._abc import Constraint


@tree.define
class FixedConstraint(Constraint):
    mask: PyTree | None = tree.field(default=None, kw_only=True)
    # TODO: support selector



from jaxtyping import Array, Float

from liblaf.peach import tree
from liblaf.peach.linalg.abc import State, Stats

type Scalar = Float[Array, ""]


@tree.define
class JaxState(State):
    pass


@tree.define
class JaxStats(Stats):
    residual_relative: Scalar = tree.array(default=None)



from collections.abc import Callable

from jaxtyping import PyTree

from liblaf.peach import tree
from liblaf.peach.functools import FunctionDescriptor, FunctionWrapper
from liblaf.peach.tree import FlatView


@tree.define
class LinearSystem(FunctionWrapper):
    matvec = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,)
    )
    """X -> X"""
    _matvec_wrapped: Callable | None = tree.field(alias="matvec")
    _matvec_wrapper: Callable | None = tree.field(default=None, init=False)

    b: PyTree = tree.field()
    b_flat = FlatView()

    rmatvec = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,)
    )
    """X -> X"""
    _rmatvec_wrapped: Callable | None = tree.field(
        default=None, alias="rmatvec", kw_only=True
    )
    _rmatvec_wrapper: Callable | None = tree.field(default=None, init=False)

    preconditioner = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,)
    )
    """X -> X"""
    _preconditioner_wrapped: Callable | None = tree.field(
        default=None, alias="preconditioner", kw_only=True
    )
    _preconditioner_wrapper: Callable | None = tree.field(default=None, init=False)

    rpreconditioner = FunctionDescriptor(
        n_outputs=1, unflatten_inputs=(0,), flatten_outputs=(0,)
    )
    """X -> X"""
    _rpreconditioner_wrapped: Callable | None = tree.field(
        default=None, alias="rpreconditioner", kw_only=True
    )
    _rpreconditioner_wrapper: Callable | None = tree.field(default=None, init=False)



from ._rosen import (
    rosen,
    rosen_grad,
    rosen_grad_and_hess_diag,
    rosen_hess_diag,
    rosen_hess_prod,
    rosen_hess_quad,
    rosen_objective,
    rosen_value_and_grad,
)

__all__ = [
    "rosen",
    "rosen_grad",
    "rosen_grad_and_hess_diag",
    "rosen_hess_diag",
    "rosen_hess_prod",
    "rosen_hess_quad",
    "rosen_objective",
    "rosen_value_and_grad",
]
